;*****WINCALLBACK***************************************************************
WINCB STRUC              ;main callback local variable space structure
WCB$1 = $
 ps PAINTSTRUCT        {}
 sinfo SCROLLINFO      {}
 tm TEXTMETRIC         {}
 PADADD        DWORD,  ?
 iVertPos      DWORD   ?
 x             DWORD   ?
 y             DWORD   ?
 ulScrollLines DWORD   ?
 column1       DWORD   ?
 column2       DWORD   ?
 column3       DWORD   ?
 szBuffer BYTE 120 DUP (?)
WCB$2 = $
 OEVESI  DWORD  ?      ;old entry value of ESI
 OEVEDI  DWORD  ?      ;old entry value of EDI
 OEVEBP  DWORD  ?      ;old entry value of EBP
 OEVEBX  DWORD  ?      ;old entry value of EBX
 RETADR  DWORD  ?      ;return address
 WHWD    DWORD  ?      ;window handle
 MSGNUM  DWORD  ?      ;message number
 WPARM   DWORD  ?      ;wParam
 LPARM   DWORD  ?      ;lParam
WINCB ENDS

 .DATA

 SJTNT1 LABEL WORD       ;start of message number table within
                         ;search jump table (SJT)
 SJTabBegin              ;build search/jump table
  SJTabEntry WM_DESTROY,DESTROY_WM             ;message/address table entry
  SJTabEntry WM_CREATE,CREATE_WM
  SJTabEntry WM_SIZE,SIZE_WM
  SJTabEntry WM_SETTINGCHANGE,SETTINGCHANGE_WM ;system wide setting change
  SJTabEntry WM_HSCROLL,HSCROLL_WM             ;horizontal scroll
  SJTabEntry WM_KEYDOWN,KEYDOWN_WM             ;keyboard scroll
  SJTabEntry WM_VSCROLL,VSCROLL_WM             ;vertical scroll
  SJTabEntry WM_MOUSEWHEEL,MOUSEWHEEL_WM       ;mousewheel scroll
  SJTabEntry WM_PAINT,PAINT_WM

 SJTNE1 EQU ($-SJTNT1)/WORD ;number of table entries
 SJTAT1 LABEL WORD       ;start of address table within search jump table (SJT)

 SJTabEnd SJTMIN,\       ;lowest entry value
          SJTMAX,\       ;highest entry value
          CREATE_WM      ;address for entry offset

 .CODE
                         ;upon entry, the stack contains the following:
                         ;ESP stack pointer -->return address
                         ;                     window handle
                         ;                     message number
                         ;                     wParam
                         ;                     lParam
WINCALLBACK:
 MOV EAX,[ESP+2*DWORD]   ;window message number from stack

 CMP EAX,SJTMAX          ;SJTMAX=highest window msg num with a user handler
 JA DefWindowProcA       ;immediately jump to default handler

 MOV EDX,SJTNE1-1        ;number of entries in window msg number table minus 1
 MOV ECX,@ SJTNT1        ;window message number table address

 .REPEAT                 ;search table for message number
   CMP AX,[ECX+WORD*EDX] ;search last entry first
   JZ @F                 ;breakout if message number found in table
  DEC EDX                ;update counter
 .UNTIL SIGN?            ;exit loop on negative counter value

 JMP DefWindowProc       ;message not found--immediate jump to default handler

@@:                      ;window message number has been found
 RPUSHIT ESI,EDI,EBP,EBX ;save essential windows registers
 MOVSX ECX,W[SJTAT1+WORD*EDX]
 MOV EBX,[ESP+7*DWORD]   ;EBX=wParam
 MOV ESI,[ESP+5*DWORD]   ;ESI=windows handle
 ADD ECX,@ CREATE_WM     ;add offset address
 MOV EDX,[ESP+8*DWORD]   ;EDX=lParam
 XOR EBP,EBP             ;EBP=0, valuable to PUSH EBP in handler routines
 SUB ESP,WCB$2-WCB$1     ;make local variable space on stack

                         ;after the above instructions, the registers and stack
                         ;contain the following useful values:
                         ;ESI=window handle
                         ;EAX=window message number
                         ;EBX=wParam
                         ;EDX=lParam
                         ;EBP=0
                         ;now, ESP stack pointer -->WINCB structure
                         ;                          old entry value of ESI
                         ;                          old entry value of EDI
                         ;                          old entry value of EBP
                         ;                          old entry value of EBX
                         ;                          return addr
                         ;                          windows handle
                         ;                          message number
                         ;                          wParam
                         ;                          lParam

 JMP ECX                 ;jump to message handler routine

DEFWINRET:               ;return point to default winproc if needed
 ADD ESP,WCB$2-WCB$1     ;recover local variable space on stack
 POPIT ESI,EDI,EBP,EBX   ;restore old entry values of essential registers
 JMP DefWindowProcA      ;for default message processing

MAINCBRET:               ;main callback return--welcome home
 ADD ESP,WCB$2-WCB$1     ;recover main callback local variable space on stack
 POPIT ESI,EDI,EBP,EBX   ;restore old entry values of essential registers
 RET 4*DWORD             ;balance stack upon returning to windows

;*****END OF MAIN CALLBACK******************************************************
